# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
Database class. Connects to MongoDB and abstracts all communication with it.
"""

import json
import time
import uuid

import mysql.connector


class DatabaseV1(object):
    """
    Database class. Connects to MongoDB and abstracts all communication with it.
    """

    def __init__(self,
                 host='127.0.0.1',
                 port=3306,
                 username=None,
                 password=None):
        self.params = {
            'host': host,
            'port': port,
            'user': username,
            'password': password,
            'autocommit': True
        }
        self.__init_db()
        self.token_lifespan = 24 * 3600

    def close(self):
        pass

    def set_setting(self, key, value):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = '''INSERT INTO moirai.settings (`key`, `value`)
                   VALUES (%s, %s) ON DUPLICATE KEY
                   UPDATE `key` = values(`key`), `value` = values(`value`)'''
        cur.execute(query, (key, json.dumps(value)))
        cur.close()
        cnx.close()

    def get_setting(self, key):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = 'SELECT `key`, `value` FROM moirai.settings WHERE `key` = %s'
        cur.execute(query, (key, ))
        r = [json.loads(value) for (key, value) in cur]
        return r[0] if len(r) > 0 else None
        cur.close()
        cnx.close()

    def verify_token(self, token):
        now = time.time()
        span = self.token_lifespan
        ts = self.get_setting('tokens')
        vs = [t for t in ts if t['token'] == token and now - t['time'] < span]
        if len(vs) > 0:
            ts = [t for t in ts if t['token'] != token]
            ts.append({'token': token, 'time': now})
            self.set_setting('tokens', ts)
            return True
        return False

    def generate_token(self):
        t = {'token': uuid.uuid4().hex, 'time': time.time()}
        ts = self.get_setting('tokens') or []
        ts += [t]
        ts = [t for t in ts if time.time() - t['time'] < self.token_lifespan]
        self.set_setting('tokens', ts)
        return t['token']

    def save_test_sensor_value(self, test, sensor, value, time, start_time):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = '''INSERT INTO moirai.sensor_values
                   (sensor, value, time, start_time, test)
                   VALUES (%s, %s, %s, %s, %s)'''
        cur.execute(query, (sensor, value, time, start_time, test))
        cur.close()
        cnx.close()

    def list_test_data(self):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = 'SELECT DISTINCT test, start_time FROM moirai.sensor_values'
        cur.execute(query)
        r = [{'name': test, 'date': start_time} for (test, start_time) in cur]
        cur.close()
        cnx.close()
        return r

    def get_test_data(self, test, start_time, skip=0):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = '''SELECT `sensor`, `time`, `value` FROM moirai.sensor_values
                   WHERE test = %s AND start_time = %s ORDER BY time LIMIT 1000000 OFFSET %s'''
        cur.execute(query, (test, start_time, skip))
        r = [{
            'sensor': sensor,
            'time': time,
            'value': value
        } for (sensor, time, value) in cur]
        cur.close()
        cnx.close()
        return r

    def get_filtered_test_data(self, test, start_time, sensors):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = '''SELECT `time`, `value` FROM moirai.sensor_values
                   WHERE test = %s AND start_time = %s AND sensor = %s'''
        result = []
        for sensor in sensors:
            s = {'sensor': sensor, 'time': [], 'values': []}
            cur.execute(query, (test, start_time, sensor))
            s['time'], s['values'] = zip(*cur)
            result.append(s)
        cur.close()
        cnx.close()
        return result

    def remove_test(self, test):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        query = '''DELETE FROM moirai.sensor_values 
                   WHERE `test` = %s AND `start_time` = %s'''
        if isinstance(test, list):
            d = [(t['test'], t['start_time']) for t in test]
            cur.executemany(query, d)
        else:
            cur.execute(query, (test['test'], test['start_time']))
        cur.close()
        cnx.close()

    def dump_database(self):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        cur.execute('SELECT `key`, `value` FROM moirai.settings')
        settings = [{
            'key': key,
            'value': json.loads(value)
        } for (key, value) in cur]
        cur.execute('''SELECT `sensor`, `value`, `time`, `start_time`, `test` 
                       FROM moirai.sensor_values''')
        sensor_values = [{
            'sensor': sensor,
            'value': value,
            'time': time,
            'start_time': start_time,
            'test': test
        } for (sensor, value, time, start_time, test) in cur]
        cur.close()
        cnx.close()
        return settings, sensor_values

    def restore_database(self, settings, test_sensor_values):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        cur.execute('DROP DATABASE IF EXISTS `moirai`')
        self.__init_db()
        query = '''INSERT INTO moirai.settings (`key`, `value`)
                   VALUES (%s, %s) ON DUPLICATE KEY
                   UPDATE `key` = values(`key`), `value` = values(`value`)'''
        data = [(s['key'], json.dumps(s['value'])) for s in settings]
        cur.executemany(query, data)

        query = '''INSERT INTO moirai.sensor_values 
                    (`sensor`, `value`, `time`, `start_time`, `test`)
                   VALUES (%s, %s, %s, %s, %s)'''
        data = [(s['sensor'], s['value'], s['time'], s['start_time'],
                 s['test']) for s in test_sensor_values]
        for d in (data[i:i + 100] for i in range(0, len(data), 100)):
            cur.executemany(query, d)
        cur.close()
        cnx.close()

    def __cnx(self):
        return mysql.connector.connect(**self.params)

    def __init_db(self):
        cnx = self.__cnx()
        cur = cnx.cursor(True)
        cur.execute('SET @@local.net_read_timeout=3600;')
        cur.execute(
            'CREATE SCHEMA IF NOT EXISTS `moirai` DEFAULT CHARACTER SET utf8')
        cur.execute('USE moirai')
        cur.execute('''CREATE TABLE IF NOT EXISTS `moirai`.`settings`
                       (`id` INT NOT NULL AUTO_INCREMENT,
                       `value` LONGTEXT NULL,
                       `key` VARCHAR(100) NOT NULL,
                       PRIMARY KEY (`id`), UNIQUE INDEX `key_UNIQUE` (`id` ASC),
                       UNIQUE INDEX `key_idx` (`key` ASC))
                       ENGINE = InnoDB DEFAULT CHARACTER SET = utf8''')
        cur.execute('''CREATE TABLE IF NOT EXISTS `moirai`.`sensor_values` 
                       ( `id` INT NOT NULL AUTO_INCREMENT,
                       `sensor` VARCHAR(100) NOT NULL, `value` DOUBLE NOT NULL,
                       `time` FLOAT NOT NULL, `start_time` DATETIME NOT NULL,
                       `test` VARCHAR(100) NOT NULL, PRIMARY KEY (`id`),
                       UNIQUE INDEX `id_UNIQUE` (`id` ASC), INDEX `sensor_idx`
                       USING BTREE (`sensor` ASC),
                       INDEX `time_idx` (`time` ASC),
                       INDEX `start_time_idx` (`start_time` ASC),
                       INDEX `test_idx` (`test` ASC))
                       ENGINE = InnoDB DEFAULT CHARACTER SET = utf8''')
        cur.close()
        cnx.close()



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import time


class AbstractProcessHandler(object):
    """
    Base class for modules ProcessHandlers. Defines the API used by two
    processes to communicate and execute code asynchronously.
    """

    def __init__(self, name, pipe):
        print('Starting %s...' % name)
        self._last_message = time.time() + 60
        self._pipes = {}
        self._pname = name  # Printable name
        self.sleep = True
        self.quitting = False
        self.set_pipe('parent', pipe)

    def set_sleep(self, sleep):
        """
        Set's the sleep property.
        """
        self.sleep = sleep

    def pipe_for(self, name):
        """
        Returns the pipe used to communicate with process `name`.
        """
        return self._pipes.get(name, None)

    def set_pipe(self, name, pipe):
        """
        Sets the pipe used to communicate with process `name`.
        """
        if pipe:
            self._pipes[name] = pipe
        else:
            del self._pipes[name]

    def pipes(self):
        """
        Returns a list of all pipes registered.
        """
        return list(self._pipes)

    def send_command(self, destination, cmd, args):
        """
        Sends a command `cmd` to the process named `destination` with
        arguments `args`
        """
        pipe = self.pipe_for(destination)
        if pipe:
            pipe.send((cmd, args))
        else:
            print("No pipe for %s" % destination)

    def read_pipe(self, name, blocking=False):
        """
        Read from pipe `name`.
        """
        pipe = self.pipe_for(name)
        try:
            if blocking or (not blocking and pipe.poll()):
                self._last_message = time.time()
                return pipe.recv()
            else:
                return (None, None)
        except EOFError:
            sname = self._pname
            print('Communication between %s and %s is closed!' % (name, sname))
            del self._pipes[name]
            return (None, None)

    def process_command(self, sender, cmd, args):
        """
        Abstract method to be implemented by classes extending this class.
        """
        pass

    def _process_command(self, name):
        cmd, args = self.read_pipe(name)
        if cmd == 'quit':
            self.quitting = True
            self.send_command(name, 'ok', None)
            return 'quit'
        elif cmd == 'close':
            pipe = self.pipe_for(name)
            pipe.send(('ok', None))
            self.set_pipe(name, None)
            if not self.pipes():
                return 'quit'
        elif cmd == 'connect':
            print('Connected %s to %s' % (args[0], self._pname))
            self.set_pipe(*args)
            self.send_command(name, 'ok', None)
        elif cmd == 'alive':
            self.send_command(name, 'alive', None)
        else:
            self.process_command(name, cmd, args)

    def request_connection(self, pkg_from, pkg_to):
        """
        Ask the main process for a pipe with process `pkg_to`.
        """
        self.send_command('parent', 'connect', (pkg_from, pkg_to))
        answer, pipe = self.read_pipe('parent', blocking=True)
        if answer == 'ok':
            self.set_pipe(pkg_to, pipe)
        else:
            raise RuntimeError(
                'Can not connect %s with %s' % (pkg_from, pkg_to))

    def loop(self):
        """
        Main loop of the process, to be implemented by the class that inherits
        this class.
        """
        pass

    def run(self):
        """
        Run loop of this process.
        """
        while True:
            if self.sleep and time.time() - self._last_message > 1:
                time.sleep(0.5)
            for name in self.pipes():
                result = self._process_command(name)
                if result == 'quit':
                    print('Shutting down %s...' % self._pname)
                    self.quit()
                    for name2 in self.pipes():
                        self.set_pipe(name2, None)
                    return
            self.loop()

    def quit(self):
        """
        Cleaning code before quitting.
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
Developer's note:
This script starts the modules as processes for multithreading purposes.
Multithreading in python is broken, so separated processes are necessary. Don't
try to revert to threads. But spawning a new process is made by clonning the
current one. That means that all variables a stuff will be copied (kinda).
Also, when a process spawns another, they all are part of the same process
groupid. When you press CTRL+C, the signal is sent to all processes in the same
groupid. To be able to cleanly exit, child processes need to know their role
and react accordingly to the signal. I do that by making the entry point of
every child process a function in this script (main(pipe, pkg)) that will
flag the process correctly, clean up resources and start the correct run loop.
The lifetime thus becomes:
 - This script starts
 - This script spawns a process
 -- Child process runs `main()` function
 -- Child's `processes` is cleared and `process_type` set to 'child'
 -- Child process enters package main loop
 - User hits CTRL+C, both process receives SIGINT
 -- Child process ignores SIGINT
 - Parent process handles it and sends 'quit' to all children through pipes
 -- Child process receives 'quit' command, cleans up and exits gracefully
 - Parent process waits for all children to terminate and then exits
"""

import hashlib
import os
import signal
import sys
import time
from multiprocessing import Pipe, Process

from moirai import __version__
from moirai import decorators
from moirai.database import DatabaseV1
from moirai.installer import install

PROCESSES = {}
PS = ['webapi', 'hardware']
PROCESS_TYPE = 'main'
WEBSOCKET = None

sys.path.append(os.path.join(os.path.splitdrive(sys.executable)[0], 'opt'))


def signal_handler(*_):
    """
    Handles system signals, reacting and sending messages to spawned processes
    """
    global PS
    # Only respond to SIGINT if we're on parent process.
    # Child processes will be asked to quit.
    if PROCESS_TYPE == 'main':
        print('')
        # Ask each child process to quit
        for key in reversed(PS):
            process, pipe = PROCESSES[key]
            pipe.send(('quit', None))
            process.join()
        print('Shutting down Moirai...')
        sys.exit(0)


def spawn_process(name):
    """
    Spawns a new processes for module named `name`
    """
    pipe_main, pipe_process = Pipe()
    process = Process(target=main, args=(pipe_process, name))
    PROCESSES[name] = (process, pipe_main)
    process.start()


def init(name):
    """
    Send the INIT command to the process named `name`
    """
    global PROCESSES
    pipe = PROCESSES[name][1]
    pipe.send(('init', None))


def main(pipe, name):
    """
    This is the main function of child processes. It will flag this process
    as child and start the event loop in the correct package. Setting
    processes to None is only to keep this instance clean, as it doesn't need
    to know about the existence of other processes.
    """
    global PROCESSES, PROCESS_TYPE
    PROCESSES = None
    PROCESS_TYPE = 'child'
    if name == 'hardware':
        import moirai.hardware as pkg
    elif name == 'webapi':
        import moirai.webapi as pkg
    pkg.main(pipe)


def query_alive(name):
    """
    Checks if process `name` is answering, e.q. if it's not stale
    """
    global PROCESSES
    pipe = PROCESSES[name][1]
    if pipe.poll():
        pipe.recv()
    pipe.send(('alive', None))
    cmd, __ = pipe.recv()
    return cmd == 'alive'


def start():
    """
    Entry point of the application
    """
    global PROCESSES, PS

    if '--version' in sys.argv:
        print(__version__)
        return

    if '--help' in sys.argv:
        print(fr'Moirai version {__version__}')
        print(r'Options:')
        print('\t--version Prints version')
        print('\t--install Tries to install dependencies.')
        print(
            '\t--sudo Uses sudo to install packages. You still need write access to /opt.'
        )
        print('\t--set-password=pwd Sets the password to pwd')
        print('')
        print('\tThe installer does not behave well with PyENV')
        return

    if '--install' in sys.argv:
        install('--sudo' in sys.argv)
        return

    # Catches SIGINT (CTRL+C)
    signal.signal(signal.SIGINT, signal_handler)
    print('Starting Moirai...')
    print('To quit press CTRL+C (^C on Macs)')
    print('Logging to %s' % decorators.log_file_path())

    # Creates a processs for each module of moirai
    for process in PS:
        spawn_process(process)

    time.sleep(1)

    while not all([query_alive(p) for p in PROCESSES]):
        pass

    # parses command line arguments
    has_cmd = False
    for arg in sys.argv:
        if arg.startswith('--set-password='):
            pswd = arg.split('=')[-1]
            if not pswd:
                pswd = None
            else:
                hasher = hashlib.sha512()
                hasher.update(bytes(pswd, 'utf-8'))
                pswd = hasher.hexdigest()
            print("Setting password to %s" % pswd)
            has_cmd = True
            database = DatabaseV1()
            database.set_setting('password', pswd)
    if has_cmd:
        signal_handler(None, None)
    else:
        for process in PS:
            init(process)
        time.sleep(1)
    last_message = time.time() + 60

    while True:
        if time.time() - last_message > 1:
            time.sleep(1)
        for name in PS:
            pipe = PROCESSES[name][1]
            if pipe.poll():
                last_message = time.time()
                command, args = pipe.recv()
                if command == 'quit':
                    signal_handler(None, None)
                elif command == 'connect':
                    pkg_from, pkg_to = args
                    pipe_to = PROCESSES[pkg_to][1]
                    process1, process2 = Pipe()
                    pipe_to.send(('connect', (pkg_from, process1)))
                    status, __ = pipe_to.recv()
                    if status == 'ok':
                        pipe.send(('ok', process2))
                    else:
                        pipe.send(('error', None))
